@@ -7,9 +7,10 @@ from decimal import Decimal
 
 from simpleeval import simple_eval
 
-from trytond.model import ModelSQL, ModelView, sequence_ordered, fields
+from trytond.model import ModelSQL, ModelView, fields, sequence_ordered
+from trytond.modules.currency.fields import Monetary
 from trytond.pool import Pool, PoolMeta
-from trytond.pyson import If, Eval
+from trytond.pyson import Eval, If
 from trytond.tools import decistmt
 from trytond.transaction import Transaction
 
@@ -47,8 +48,17 @@ class Statement(metaclass=PoolMeta):
             for rule in rules:
                 keywords = rule.match(origin)
                 if keywords:
+                    origin.keywords = keywords
                     yield from rule.apply(origin, keywords)
                     break
+        self.origins = self.origins
+        self.save()
+
+
+class StatementOrigin(metaclass=PoolMeta):
+    __name__ = 'account.statement.origin'
+
+    keywords = fields.Dict(None, "Keywords")
 
 
 class StatementRule(sequence_ordered(), ModelSQL, ModelView):
@@ -64,28 +74,23 @@ class StatementRule(sequence_ordered(), ModelSQL, ModelView):
             If(Eval('company'),
                 ('company', '=', Eval('company')),
                 ()),
-            ],
-        depends=['company'])
-    amount_low = fields.Numeric(
-        "Amount Low",
-        digits=(16, Eval('currency_digits', 2)),
+            ])
+    amount_low = Monetary(
+        "Amount Low", currency='currency', digits='currency',
         domain=[If(Eval('amount_high'),
                 ['OR',
                     ('amount_low', '=', None),
                     ('amount_low', '<=', Eval('amount_high')),
                     ],
-                [])],
-        depends=['currency_digits', 'amount_high'])
-    amount_high = fields.Numeric(
-        "Amount High",
-        digits=(16, Eval('currency_digits', 2)),
+                [])])
+    amount_high = Monetary(
+        "Amount High", currency='currency', digits='currency',
         domain=[If(Eval('amount_low'),
                 ['OR',
                     ('amount_high', '=', None),
                     ('amount_high', '>=', Eval('amount_low')),
                     ],
-                [])],
-        depends=['currency_digits', 'amount_low'])
+                [])])
     description = fields.Char("Description",
         help="The regular expression the description is searched with.\n"
         "It may define the groups named:\n"
@@ -96,15 +101,13 @@ class StatementRule(sequence_ordered(), ModelSQL, ModelView):
     lines = fields.One2Many(
         'account.statement.rule.line', 'rule', "Lines")
 
-    currency_digits = fields.Function(
-        fields.Integer("Currency Digits"),
-        'on_change_with_currency_digits')
+    currency = fields.Function(fields.Many2One(
+            'currency.currency', "Currency"),
+        'on_change_with_currency')
 
     @fields.depends('journal')
-    def on_change_with_currency_digits(self, name=None):
-        if self.journal:
-            return self.journal.currency.digits
-        return None
+    def on_change_with_currency(self, name=None):
+        return self.journal.currency if self.journal else None
 
     def match(self, origin):
         keywords = {}
@@ -158,13 +161,11 @@ class StatementRuleInformation(sequence_ordered(), ModelSQL, ModelView):
     boolean = fields.Boolean("Boolean",
         states={
             'invisible': Eval('key_type') != 'boolean',
-            },
-        depends=['key_type'])
+            })
     char = fields.Char("Char",
         states={
             'invisible': Eval('key_type') != 'char',
             },
-        depends=['key_type'],
         help="The regular expression the key information is searched with.\n"
         "It may define the groups named:\n"
         "party, bank_account, invoice.")
@@ -172,8 +173,7 @@ class StatementRuleInformation(sequence_ordered(), ModelSQL, ModelView):
         'get_selections', "Selection",
         states={
             'invisible': Eval('key_type') != 'selection',
-            },
-        depends=['key_type'])
+            })
 
     key_type = fields.Function(
         fields.Selection('get_key_types', "Key Type"),
@@ -253,8 +253,7 @@ def _add_range(cls, name, type_, string):
                     [])],
             states={
                 'invisible': Eval('key_type') != name,
-                },
-            depends=['key_type', high_name]))
+                }))
     setattr(cls, high_name,
         type_("%s High" % string,
             domain=[If(Eval(low_name),
@@ -265,8 +264,7 @@ def _add_range(cls, name, type_, string):
                     [])],
             states={
                 'invisible': Eval('key_type') != name,
-                },
-            depends=['key_type', low_name]))
+                }))
 
 
 _add_range(StatementRuleInformation, 'integer', fields.Integer, "Integer")
@@ -288,7 +286,7 @@ class StatementRuleLine(sequence_ordered(), ModelSQL, ModelView):
         context={
             'company': Eval('company', -1),
             },
-        depends=['company'],
+        depends={'company'},
         help="Leave empty to use the group named 'party' "
         "from the regular expressions.")
     account = fields.Many2One(
@@ -300,7 +298,6 @@ class StatementRuleLine(sequence_ordered(), ModelSQL, ModelView):
         states={
             'readonly': ~Eval('company'),
             },
-        depends=['company'],
         help="Leave empty to use the party's receivable or payable account.\n"
         "The rule must have a company to use this field.")
 
@@ -315,8 +312,7 @@ class StatementRuleLine(sequence_ordered(), ModelSQL, ModelView):
 
     @fields.depends('rule', '_parent_rule.company')
     def on_change_with_company(self, name=None):
-        if self.rule and self.rule.company:
-            return self.rule.company.id
+        return self.rule.company if self.rule else None
 
     def get_line(self, origin, keywords, **context):
         pool = Pool()
@@ -327,17 +323,25 @@ class StatementRuleLine(sequence_ordered(), ModelSQL, ModelView):
         currency = origin.statement.journal.currency
         amount = currency.round(simple_eval(decistmt(self.amount), **context))
         party = self._get_party(origin, keywords)
-        invoice = self._get_invoice(origin, keywords)
+        related_to = list(
+            filter(None, self._get_related_to(
+                    origin, keywords, party=party, amount=amount)))
+        if len(related_to) == 1:
+            related_to, = related_to
+        else:
+            related_to = None
+        related_to_party = self._get_party_from(related_to)
 
-        if invoice and party and invoice.party != party:
+        if related_to_party and party and related_to_party != party:
             return
-        if invoice and not party:
-            party = invoice.party
+        if related_to_party and not party:
+            party = related_to_party
 
         account = self.account
         if not account:
-            if invoice:
-                account = invoice.account
+            related_to_account = self._get_account_from(related_to)
+            if related_to_account:
+                account = related_to_account
             elif party:
                 with Transaction().set_context(date=origin.date):
                     if amount > Decimal(0):
@@ -363,12 +367,13 @@ class StatementRuleLine(sequence_ordered(), ModelSQL, ModelView):
         line.date = origin.date
         line.party = party
         line.account = account
-        line.invoice = invoice
+        line.related_to = related_to
         return line
 
     def _get_party(self, origin, keywords):
         pool = Pool()
         Party = pool.get('party.party')
+        Line = pool.get('account.statement.line')
         try:
             AccountNumber = pool.get('bank.account.number')
         except KeyError:
@@ -386,18 +391,65 @@ class StatementRuleLine(sequence_ordered(), ModelSQL, ModelView):
                     number, = numbers
                     if number.account.owners:
                         party = number.account.owners[0]
+                else:
+                    lines = Line.search([
+                            ('statement.state', 'in', ['validated', 'posted']),
+                            ('origin.keywords.bank_account', '=',
+                                bank_account),
+                            ('party', '!=', None),
+                            ],
+                        order=[('date', 'DESC')], limit=1)
+                    if lines:
+                        line, = lines
+                        party = line.party
             elif keywords.get('party'):
                 parties = Party.search(
                     [('rec_name', 'ilike', keywords['party'])])
                 if len(parties) == 1:
                     party, = parties
+                else:
+                    lines = Line.search([
+                            ('statement.state', 'in', ['validated', 'posted']),
+                            ('origin.keywords.party', '=', keywords['party']),
+                            ('party', '!=', None),
+                            ],
+                        order=[('date', 'DESC')], limit=1)
+                    if lines:
+                        line, = lines
+                        party = line.party
         return party
 
-    def _get_invoice(self, origin, keywords):
+    def _get_related_to(self, origin, keywords, party=None, amount=0):
+        return {self._get_invoice(origin, keywords, party=party)}
+
+    def _get_party_from(self, related_to):
+        pool = Pool()
+        Invoice = pool.get('account.invoice')
+        if isinstance(related_to, Invoice):
+            return related_to.party
+
+    def _get_account_from(self, related_to):
+        pool = Pool()
+        Invoice = pool.get('account.invoice')
+        if isinstance(related_to, Invoice):
+            return related_to.account
+
+    def _get_invoice(self, origin, keywords, party=None):
         pool = Pool()
         Invoice = pool.get('account.invoice')
         if keywords.get('invoice'):
-            invoices = Invoice.search([('rec_name', '=', keywords['invoice'])])
+            domain = [
+                ('rec_name', '=', keywords['invoice']),
+                ('company', '=', origin.company.id),
+                ('currency', '=', origin.currency.id),
+                ('state', '=', 'posted'),
+                ]
+            if party:
+                domain.append(['OR',
+                        ('party', '=', party.id),
+                        ('alternative_payees', '=', party.id),
+                        ])
+            invoices = Invoice.search(domain)
             if len(invoices) == 1:
                 invoice, = invoices
                 return invoice
